<?php
App::uses('EvShopAppModel', 'EvShop.Model');
App::uses('CakeNumber', 'Utility');
/**
 * Product Model
 *
 * @property Brand $Brand
 * @property Variant $Variant
 * @property Category $Category
 */
class Product extends EvShopAppModel {

	public $actsAs = array(
		'Routable.Routable' => array(
			'config' => 'EvShop',
			'alias' => 'product/:displayField'
		),
		'MetaData.Meta',
		'EvTemplates.Template' => array(
			'formInject' => true,
			'restrictTo' => 'Product'
		),
		'EvShop.ProductCopyable' => [
			'recurse_level' => 2,
		],
	);

	/**
	 * Display field
	 *
	 * @var string
	 */
	public $displayField = 'name';

	/**
	 * A cached copy of the tax levels to avoid hitting the DB too many times
	 * @var array
	 */
	protected $_taxLevels = null;

	/**
	 * Validation rules
	 *
	 * @var array
	 */
	public $validate = array(
		'name' => array(
			'notEmpty' => array(
				'rule' => array('notBlank'),
				'message' => 'A product name must be entered'
			),
		),
	);

	//The Associations below have been created with all possible keys, those that are not needed can be removed

	/**
	 * belongsTo associations
	 *
	 * @var array
	 */
	public $belongsTo = array(
		'Brand' => array(
			'className' => 'EvShop.Brand'
		),
	);

	/**
	 * hasMany associations
	 *
	 * @var array
	 */
	public $hasMany = array(
		'Variant' => array(
			'className' => 'EvShop.Variant',
			'cascade' => true,
			'dependent' => true
		),
		'CategoriesProduct' => array(
			'className' => 'EvShop.CategoriesProduct',
			'order' => 'CategoriesProduct.sequence ASC'
		),
	);

	public $hasAndBelongsToMany = [
		'ProductAttribute' => [
			'className' => 'EvShop.ProductAttribute',
			'joinTable' => 'ev_shop_product_attributes_products'
		]
	];

	/**
	 * image slots
	 */
	public $imageSlots = array(
		'listing' => array(
			'slots' => 1,
			'fields' => false
		)
	);

	/**
	 * redefine constructor to check for EvTax
	 */
	public function __construct($id = false, $table = null, $ds = null) {
		if (CakePlugin::loaded('EvTax')) {
			$this->actsAs['EvTax.Taxable'] = array(
				'formInject' => true
			);

			$this->belongsTo['TaxLevel'] = array(
				'className' => 'EvTax.TaxLevel'
			);
		}

		parent::__construct($id, $table, $ds);
	}

	/**
	 * before we save anything process the options into variants
	 * also process the category array
	 * also process the variant images
	 */
	public function beforeBeforeSave($data, $options) {
		$Variant = EvClassRegistry::init('EvShop.Variant');

		if (empty($data['CurrentOptions'])) {
			$data['CurrentOptions'] = array();
		}
		if (empty($data['Options'])) {
			$data['Options'] = array();
		}

		$data['Options'] = array_map(
			function ($options) {
				return array_combine(
					$options,
					$options
				);
			},
			$data['Options']
		);

		//Match the current options array with the selected options array
		$data['CurrentOptions'] = array_map(
			function ($options) {
				return array_combine(
					$options,
					$options
				);
			},
			$data['CurrentOptions']
		);

		// Take a copy of any current Variant data so it doesn't get overwritten later
		$originalVariants = array();
		if (isset($data['Variant'])) {
			$originalVariants = $data['Variant'];
		}

		//Only recalculate the variants if there are options to parse or if a new record is being saved.
		if (!empty($data['Options']) || empty($data[$this->alias][$this->primaryKey])) {
			if (!empty($data[$this->alias][$this->primaryKey])) {
				$this->variantsToDelete = $Variant->processDeletedOptions($data['CurrentOptions'], $data['Options'], $data[$this->alias][$this->primaryKey]);
			}

			if (!empty($data[$this->alias])) {
				//We need to process options for new products so that products without options get a variant.
				$data['Variant'] = $Variant->processNewOptions($data['CurrentOptions'], $data['Options'], $data[$this->alias]);
				unset($data['Options']);
			}
		}

		// Merge the original into the list if it is not on the list to delete
		// This stops old redundant variants being recreated as variant save happens after product afterSave deletes the variants from variantsToDelete
		$idsToDelete = !empty($this->variantsToDelete) ? array_keys($this->variantsToDelete) : [];
		if (!empty($originalVariants)) {
			foreach ($originalVariants as $variant) {
				if (!empty($variant['id']) && !in_array($variant['id'], $idsToDelete)) {
					$data['Variant'][] = $variant;
				}
			}
		}

		// process the categories
		$Category = EvClassRegistry::init('EvShop.Category');
		$this->categoriesToDelete = $Category->processDeletedCategories($data['CategoriesProduct']);

		// process the variant images
		if (isset($this->variantImageSlots)) {
			$VariantImage = EvClassRegistry::init('EvShop.VariantImage');
			$data = $VariantImage->processVariantImageOptions($data, $this->variantImageSlots);
		}

		return $data;
	}

	/**
	 * afterSave - delete any variants to be deleted
	 *
	 */
	public function afterSave($created, $options = array()) {
		parent::afterSave($created, $options);

		if (! empty($this->variantsToDelete)) {
			$Variant = EvClassRegistry::init('EvShop.Variant');

			$Variant->deleteAll(
				array(
					'Variant.id' => array_keys($this->variantsToDelete)
				)
			);
		}

		if (! empty($this->categoriesToDelete)) {
			$CategoriesProduct = EvClassRegistry::init('EvShop.CategoriesProduct');

			$CategoriesProduct->deleteAll(
				array(
					'CategoriesProduct.id' => $this->categoriesToDelete
				)
			);
		}

		// dispatch product saved event
		$this->getEventManager()->dispatch(
			new CakeEvent('EvShop.Model.Product.saved', $this, array(
				'created' => $created,
				'product' => $this->data['Product']
			))
		);

		$this->clearCache($this->id);
	}

	/**
	 * afterFind - Calculate tax prices for each variant
	 *
	 */

	public function afterFind($results, $primary = false) {
		if (!empty($results)) {
			if (CakePlugin::loaded('EvTax')) {
				$results = $this->calculateTaxPrices($results);
			}
		}

		return parent::afterFind($results, $primary);
	}

	/**
	 * Called after every deletion operation.
	 *
	 * @return void
	 * @link http://book.cakephp.org/2.0/en/models/callback-methods.html#afterdelete
	 */
	public function afterDelete() {
		$this->CategoriesProduct->deleteAll(['CategoriesProduct.product_id' => $this->id], false);
	}

	/**
	 * readForEdit - bring out the variants and calculate the options
	 *
	 */
	public function readForEdit($id, $query = array()) {
		$query['contain']['CategoriesProduct'] = array('Category');

		if (!empty(Configure::read('EvShop.showAttributes'))) {
			$query['contain']['ProductAttribute'] = [];
		}

		// get product and product category data
		$data = parent::readForEdit($id, $query);

		if (! empty($data['Product'])) {
			// re-align the category products array
			$data['CategoriesProduct'] = Hash::combine(
				$data['CategoriesProduct'],
				'{n}.category_id',
				'{n}'
			);

			// get variants out for the product, containing pricing, ready for
			// in-page editing through the admin
			$data['Variant'] = $this->Variant->readForManage($id);

			$data['Options'] = [];

			// assign array of active options to auto-check items under the Options tab
			if (! empty($data['Variant'])) {
				$data['Options'] = Hash::combine(
					$data['Variant'],
					'{n}.Option.{n}.id',
					'{n}.Option.{n}.id',
					'{n}.Option.{n}.option_group_id'
				);
			}
		}

		return $data;
	}

	/**
	 * readForView - Do the Variants separately so the callbacks will be run
	 *
	 * @param integer $id ID of row to edit
	 * @param array $params The db query array - can be used to pass in additional parameters such as contain
	 * @return array
	 */
	public function readForView($id, $query = array()) {
		$query['contain']['CategoriesProduct'] = array('Category');

		if (Configure::read('EvShop.showBrands')) {
			$query['contain'][] = 'Brand';
		}

		$data = parent::readForView($id, $query);

		if (empty($data)) {
			return false;
		}

		// perform this on it's own rather then contain so the inventory behaviours / cakllbacks can be triggered
		$data['Variant'] = $this->Variant->readForProductView($id);

		if (! empty($data['Variant'])) {
			$data['Options'] = Hash::combine($data['Variant'], '{n}.Option.{n}.id', '{n}.Option.{n}.id', '{n}.Option.{n}.option_group_id');
		}

		if (! empty($data['CategoriesProduct'])) {
			$data['CategoriesProduct'] = Hash::combine($data['CategoriesProduct'], '{n}.category_id', '{n}');
		}

		$data = $this->defaultMetaDescription($data);

		return $data;
	}

	/**
	 * unless the meta description has been overridden
	 * set seo product meta description
	 *
	 * @param 	array 	$data 	Array of product data
	 * @return 	array 	$data 	Updated product data array with updated meta description
	 */
	public function defaultMetaDescription($data) {
		if (! isset($data['MetaData']['description']) || empty($data['MetaData']['description'])) {
			$variantCount = count($data['Variant']);
			$siteTitle = Configure::read('SiteSetting.general.site_title');

			$currencies = Configure::read('currencies');

			if ($variantCount > 1) {
				if (Configure::read('EvShop.displayTaxAsVat')) {
					$prices = Hash::extract($data, 'Variant.{n}.VariantPricing.{n}[currency_id=' . CakeSession::read('EvCurrency.currencyId') . '][price_incTax>0].price_incTax');
				} else {
					$prices = Hash::extract($data, 'Variant.{n}.VariantPricing.{n}[currency_id=' . CakeSession::read('EvCurrency.currencyId') . '][price>0].price');
				}

				if (!empty($prices)) {
					sort($prices);

					$price = array_shift($prices);
					$price = CakeNumber::currency(
						$price,
						$currencies[CakeSession::read('EvCurrency.currencyId')]
					);

					$description = "Order " . $data['Product']['name'] . " from only " . $price . " from " . $siteTitle . " now!";
				}
			} elseif ($variantCount === 1) {
				if (!isset($data['Variant'][0]['VariantPricing'][CakeSession::read('EvCurrency.currencyId')])) {
					$price = 0;
				} else {
					$pricing = $data['Variant'][0]['VariantPricing'][CakeSession::read('EvCurrency.currencyId')];

					$price = $pricing['price'];
					if (Configure::read('EvShop.displayTaxAsVat') && isset($pricing['price_incTax'])) {
						$price = $pricing['price_incTax'];
					}
				}

				$price = CakeNumber::currency(
					$price,
					$currencies[CakeSession::read('EvCurrency.currencyId')]
				);

				$description = "Order " . $data['Product']['name'] . " for only " . $price . " from " . $siteTitle . " now!";
			}

			if (! empty($description)) {
				$data['MetaData']['description'] = $description;
			}
		}

		return $data;
	}

	/**
	 * Calculate the prices with tax for each variant
	 *
	 * @param array $result The results from the find
	 * @return array $result Updated variant pricing to include tax prices
	 */

	public function calculateTaxPrices($result) {
		$TaxLevelModel = EvClassRegistry::init('EvTax.TaxLevel');
		$VariantPricingModel = EvClassRegistry::init('EvShop.VariantPricing');

		// Cache the tax levels in the model - should be safe in the same call.
		if (empty($this->_taxLevels)) {
			$this->_taxLevels = $TaxLevelModel->find('all', array(
				'conditions' => array(
					'is_active' => 1
				)
			));
			$this->_taxLevels = Hash::combine($this->_taxLevels, '{n}.TaxLevel.id', '{n}');
		}

		if (! empty($result)) {
			foreach ($result as $key => $item) {
				$taxLevel = array();

				if (isset($item['Product']['tax_level_id']) && $item['Product']['tax_level_id'] > 0) {

					$taxLevel = $this->_taxLevels[$item['Product']['tax_level_id']];

					// As well as the variants add tax to the product 'summary' prices
					if (isset($item['Product']['was_price_ex_vat'])) {
						if (isset($taxLevel['TaxLevel']['rate'])) {
							$result[$key]['Product']['was_price'] = $VariantPricingModel->addTaxToPrice($item['Product']['was_price_ex_vat'], $taxLevel[$TaxLevelModel->alias]['rate']);
						} else {
							$result[$key]['Product']['was_price'] = $item['Product']['was_price_ex_vat'];
						}
					}
					if (isset($item['Product']['rrp_price_ex_vat'])) {
						if (isset($taxLevel['TaxLevel']['rate'])) {
							$result[$key]['Product']['rrp_price'] = $VariantPricingModel->addTaxToPrice($item['Product']['rrp_price_ex_vat'], $taxLevel[$TaxLevelModel->alias]['rate']);
						} else {
							$result[$key]['Product']['rrp_price'] = $item['Product']['rrp_price_ex_vat'];
						}
					}

					if (isset($item['Variant']) && ! empty($item['Variant'])) {
						foreach ($item['Variant'] as $vKey => $variant) {

							if (! empty($taxLevel['TaxLevel'])) {

								if (isset($variant['VariantPricing']) && ! empty($variant['VariantPricing'])) {

									foreach ($variant['VariantPricing'] as $pricingKey => $pricing) {
										$variant['VariantPricing'][$pricingKey]['TaxLevel'] = $taxLevel['TaxLevel'];
									}
								}
							}

							if (isset($result[$key]['Variant'][$vKey])) {
								$result[$key]['Variant'][$vKey] = $VariantPricingModel->calculateTax($variant, 'Variant');
							}
						}
					}
				}
			}
		}

		return $result;
	}

/**
 * Clears the cache stored against this product
 *
 * @param int $id The product ID. If null, all of the products will be cleared.
 * @param string $key The cache key to delete. If null, all of the cached data against this product will be deleted
 * @return void
 */
	public function clearCache($id = null, $key = null) {
		$cacheName = Configure::read('EvShop.productCacheName');
		if (!empty($cacheName)) {
			if (!empty($id)) {
				if (empty($key)) {
					// Delete all product specific data
					Cache::delete('EvShop.Product.' . $id, $cacheName);
				} else {
					$cachedData = Cache::read('EvShop.Product.' . $id);
					unset($cachedData[$key]);
					if (empty($cachedData)) {
						// If it's empty delete the whole thing
						$this->clearCache($id);
					} else {
						// Else save the rest back to the cache
						Cache::write('EvShop.Product.' . $id, $cachedData, $cacheName);
					}
				}
			} else {
				Cache::clear(false, $cacheName);
			}
		}
	}

/**
 * Cache an item against a product
 *
 * @param int $id The product id to cache against
 * @param string $key The cache key
 * @param mixed $data The data to store. Must be serialisable
 * @return void
 */
	public function cache($id, $key, $data) {
		$cacheName = Configure::read('EvShop.productCacheName');
		if (!empty($cacheName)) {
			$cachedData = Cache::read('EvShop.Product.' . $id, $cacheName);
			$cachedData[$key] = $data;
			Cache::write('EvShop.Product.' . $id, $cachedData, $cacheName);
		}
	}

/**
 * Fetch cached data from a product
 *
 * @param int $id The product id to cache against
 * @param string $key The cache key. If null it will fetch the entire data stored against this product
 * @param mixed $data The data to store. Must be serialisable
 * @return void
 */
	public function fetchFromCache($id, $key = null) {
		$cacheName = Configure::read('EvShop.productCacheName');
		if (!empty($cacheName)) {
			$cachedData = Cache::read('EvShop.Product.' . $id, $cacheName);
			if (!empty($key)) {
				return isset($cachedData[$key]) ? $cachedData[$key] : null;
			}
			return $cachedData;
		}
		return null;
	}
}
